/* MAC0499 - Trabalho de Formatura Supervisionado - Compresso de udio Digital
 * Marcio Masaki Tomiyoshi */

#include "decoder.h"

/* Decodificador AudioPaK */

/* Devolve 0 para frames completos, tamanho do frame caso contrrio */
int readFrameHeader(FILE *input, int *predictor, int *k)
{
	int incompleteFrame = readInteger(input, INCOMPLETE_FRAME_BITS);
	*predictor = readInteger(input, PREDICTOR_BITS);
	*k = readInteger(input, PARAMETER_M_BITS);

	int frameSize = 0;
	if (incompleteFrame)
		frameSize = readInteger(input, MAX_FRAME_SIZE_BITS);

	return (int)frameSize;
}

/* Decodifica nmeros codificados em Golomb */
int golombDecode(FILE *input, int k)
{
	/* Bit de sinal */
	int negative = readBit(input);
	int i = k, n = 0, r = 0;

	/* Codificao unria */
	while (readBit(input) == 0)
		n++;

	/* k bits menos significativos */
	while (i > 0)
		r = r | (readBit(input) << --i);

	r += n * (1 << k);

	if (negative) r = -r;

	return r;
}

/* Testa se o cabealho do arquivo  vlido */
int isValidFileHeader(FILE *input)
{
	char temp[5];

	fscanf(input, "%4s", temp);
	temp[4] = '\0';

	return (strcmp(temp, "COMP") == 0);
}

/* Funo de decodificao AudioPaK */
void decode(char *filein, char *fileout)
{
	FILE *input = NULL, *output = NULL;

	input = openFile(filein, "rb");
	output = openFile(fileout, "wb");

	if (isValidFileHeader(input))
	{
		char temp[44];
		int frameSize, numChannels, bitsPerSample;

		/* Lemos o cabealho AudioPaK */
		frameSize = readInteger(input, MAX_FRAME_SIZE_BITS);
		numChannels = readInteger(input, MAX_NUM_CHANNELS_BITS);
		bitsPerSample = readInteger(input, MAX_BITS_PER_SAMPLE_BITS);
		readBytes(input, temp, 44);
		writeBytes(output, temp, 44);

		int encodedSamples[NUM_PREDICTOR][frameSize][numChannels];
		int incomplete, k, framePredictor, frame, channel, predictor, initialSample;

		/* Lemos o cabealho do bloco */
		incomplete = readFrameHeader(input, &framePredictor, &k);

		while (incomplete == 0)
		{
			/* Lemos a amostra de referncia do bloco */
			for (channel = 0; channel < numChannels; channel++)
			{
				initialSample = readInteger(input, bitsPerSample);
				initialSample = twoComplementToInt(initialSample, bitsPerSample / 8);
				for (predictor = 0; predictor < NUM_PREDICTOR; predictor++)
					encodedSamples[predictor][0][channel] = initialSample;
			}
			/* Lemos os residuais seguintes e calculamos os valores das amostras */
			for (frame = 1; frame < frameSize; frame++)
			{
				for (channel = 0; channel < numChannels; channel++)
					encodedSamples[framePredictor][frame][channel] = golombDecode(input, k);
				for (predictor = framePredictor; predictor > 0; predictor--)
					for (channel = 0; channel < numChannels; channel++)
						encodedSamples[predictor-1][frame][channel] = encodedSamples[predictor][frame][channel] + encodedSamples[predictor-1][frame-1][channel];
			}
			/* Escrevemos as amostras na sada */
			for (frame = 0; frame < frameSize; frame++)
				for (channel = 0; channel < numChannels; channel++)
					writeWaveSample(output, encodedSamples[0][frame][channel], bitsPerSample / 8);
			/* Lemos o prximo cabealho de bloco */
			incomplete = readFrameHeader(input, &framePredictor, &k);
		}
		/* Tratamos o ltimo bloco de forma equivalente a acima */
		frameSize = incomplete;
		for (channel = 0; channel < numChannels; channel++)
		{
			initialSample = readInteger(input, bitsPerSample);
			initialSample = twoComplementToInt(initialSample, bitsPerSample / 8);
			for (predictor = 0; predictor < NUM_PREDICTOR; predictor++)
				encodedSamples[predictor][0][channel] = initialSample;
		}
		for (frame = 1; frame < frameSize; frame++)
		{
			for (channel = 0; channel < numChannels; channel++)
				encodedSamples[framePredictor][frame][channel] = golombDecode(input, k);
			for (predictor = framePredictor; predictor > 0; predictor--)
				for (channel = 0; channel < numChannels; channel++)
					encodedSamples[predictor-1][frame][channel] = encodedSamples[predictor][frame][channel] + encodedSamples[predictor-1][frame-1][channel];
		}
		for (frame = 0; frame < frameSize; frame++)
			for (channel = 0; channel < numChannels; channel++)
				writeWaveSample(output, encodedSamples[0][frame][channel], bitsPerSample / 8);
	}
	else
		showError("Not a compressed file.\n");

	fclose(input);
	fclose(output);
}
