/* MAC0499 - Trabalho de Formatura Supervisionado - Compresso de udio Digital
 * Marcio Masaki Tomiyoshi */

#include "encoder.h"

/* Codificador AudioPaK */

/* Escreve o cabealho do arquivo AudioPaK */
void writeFileHeader(FILE *input, FILE *output, int frameSize, int numChannels, int bitsPerSample)
{
	fprintf(output, "COMP");
	writeInteger(output, frameSize, MAX_FRAME_SIZE_BITS);
	writeInteger(output, numChannels, MAX_NUM_CHANNELS_BITS);
	writeInteger(output, bitsPerSample, MAX_BITS_PER_SAMPLE_BITS);

	copyWaveHeader(input, output);
}

/* Escreve o cabealho do bloco */
/* TODO Devolve 1 se o frame  preenchido apenas por 0, 0 caso contrrio */
void writeFrameHeader(FILE *output, int predictor, unsigned int absError, int parameterM, int incompleteFrame, int frameSize)
{
	writeInteger(output, incompleteFrame && 1, INCOMPLETE_FRAME_BITS);
	writeInteger(output, predictor, PREDICTOR_BITS);	/* TODO frame s com 0 */
	writeInteger(output, parameterM, PARAMETER_M_BITS);

	if (incompleteFrame)
		writeInteger(output, frameSize, MAX_FRAME_SIZE_BITS);
}

/* Devolve o estimador com menor valor absoluto */
int calculatePredictor(unsigned int totalSize[])
{
	int predictor = 0, min = totalSize[0], l;

	for (l = 1; l < NUM_PREDICTOR; l++)
		if (totalSize[l] < min)
		{
			min = totalSize[l];
			predictor = l;
		}

	return predictor;
}

/* Estima o parmetro k do cdigo de Golomb */
int estimateM(int frameSize, int absError)
{
	if (absError == 0)
		return 0;

	int r = (int)fabs((log10(log(2)*absError/frameSize)/log10(2)));

	return r;
}

/* Faz a codificao de Golomb */
void golombEncode(FILE *output, int sample, int k)
{
	int i = k;
	int q;

	#if DEBUG == 3
	printf("%d sample: %d\n", k, sample);
	#endif

	writeBit(output, sample < 0); /* Sinal */

	/* Codificao unria */
	int absSample = abs(sample);
	q = absSample / (1 << k);

	while (q-- > 0)
		writeBit(output, 0);
	writeBit(output, 1);

	/* k bits menos significativos */
	while (i > 0)
		writeBit(output, absSample & (1 << --i));

	#if DEBUG == 3
	printf("\n");
	#endif
}

/* Funo de codificao */
void encode(char *filein, char *fileout, int frameSize)
{
	FILE *input = NULL, *output = NULL;
	WaveHeader wh;

	unsigned int numSamples, bytesPerSample, sample;
	unsigned int totalSize[NUM_PREDICTOR];

	int channel, predictor, frame, currentSample;

	wh = openWave(&input, filein);
	output = openFile(fileout, "wb");

	int encodedSamples[NUM_PREDICTOR][frameSize][wh.numChannels];

	bytesPerSample = wh.bitsPerSample / 8;
	numSamples = calculateSamplesNumber(wh);

	/* Escrevemos o cabealho AudioPaK */
	writeFileHeader(input, output, frameSize, wh.numChannels, wh.bitsPerSample);

	int frameIndex = 0;
	int currentFrame = 0;
	int lastFrame = 0;
	int lastFrameSize = -1;

	memset(totalSize, 0, sizeof(unsigned int) * NUM_PREDICTOR);
	memset(encodedSamples, 0, sizeof(int) * NUM_PREDICTOR * frameSize * wh.numChannels);

	for (currentSample = 0; currentSample < numSamples; currentSample++)
	{
		frameIndex = currentSample % frameSize;

		/* Testamos se  a primeira amostra do bloco */
		if (frameIndex == 0)
		{
			currentFrame++;
			/* Testamos se  o ltimo bloco do arquivo e calculamos o tamanho dele */
			if (currentFrame * frameSize >= numSamples)
			{
				lastFrame = 1;
				lastFrameSize = numSamples - (currentFrame - 1) * frameSize;
			}

			/* Armazenamos a amostra de referncia */
			for (channel = 0; channel < wh.numChannels; channel++)
			{
				sample = readWaveSample(input, bytesPerSample);
				for (predictor = 0; predictor < NUM_PREDICTOR; predictor++)
				{
					encodedSamples[predictor][frameIndex][channel] = sample;
					totalSize[predictor] += abs(encodedSamples[predictor][frameIndex][channel]);
				}
			}
		}
		else
		{
			for (channel = 0; channel < wh.numChannels; channel++)
			{
				/* Lemos uma amostra e calculamos os residuais */
				sample = readWaveSample(input, bytesPerSample);
				encodedSamples[0][frameIndex][channel] = sample;
				totalSize[0] += abs(encodedSamples[0][frameIndex][channel]);

				for (predictor = 1; predictor < NUM_PREDICTOR; predictor++)
				{
					encodedSamples[predictor][frameIndex][channel] =
						encodedSamples[predictor-1][frameIndex][channel] - encodedSamples[predictor-1][frameIndex-1][channel];
					totalSize[predictor] += abs(encodedSamples[predictor][frameIndex][channel]);
				}
			}
			/* Se for a ltima amostra do bloco */
			if (frameIndex == frameSize - 1 || (currentSample == numSamples - 1))
			{
				/* Escolhemos o estimador */
				int predictor = calculatePredictor(totalSize);
				/* Estimamos o parmetro de Golomb */
				int parameterM = estimateM(frameSize, totalSize[predictor]);

				/* Escrevemos o cabealho */
				writeFrameHeader(output, predictor, totalSize[predictor], parameterM, lastFrame, lastFrameSize);
				/* Escrevemos a amostra de referncia */
				for (channel = 0; channel < wh.numChannels; channel++)
					writeInteger(output, intToTwoComplement(encodedSamples[predictor][0][channel], bytesPerSample), wh.bitsPerSample);
				/* Escrevemos os residuais */
				for (frame = 1; frame < frameSize; frame++)
					for (channel = 0; channel < wh.numChannels; channel++)
						golombEncode(output, encodedSamples[predictor][frame][channel], parameterM);

				memset(totalSize, 0, sizeof(unsigned int) * NUM_PREDICTOR);
				memset(encodedSamples, 0, sizeof(int) * NUM_PREDICTOR * frameSize * wh.numChannels);
			}
		}
	}

	flushBit(output);

	fclose(input);
	fclose(output);
}
