/* MAC0499 - Trabalho de Formatura Supervisionado - Compresso de udio Digital
 * Marcio Masaki Tomiyoshi */

#include <stdio.h>
#include <string.h>

#include "io.h"

/* IMA ADPCM */

/* Nmero de bytes da palavra */
#define WORD_SIZE 4
/* Tamanho do bloco */
#define BLOCK_SIZE 2048

/* Tabela de ndices. Tem o dobro do tamanho para facilitar a consulta */
const int indexAdjustment[] = {-1, -1, -1, -1, 2, 4, 6, 8,
		-1, -1, -1, -1, 2, 4, 6, 8};
/* Tabela com os tamanhos de passo */
const int stepSize[] = {7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 19,
		21, 23, 25, 28, 31, 34, 37, 41, 45, 50,
		55, 60, 66, 73, 80, 88, 97, 107, 118, 130,
		143, 157, 173, 190, 209, 230, 253, 279, 307, 337,
		371, 408, 449, 494, 544, 598, 658, 724, 796, 876,
		963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066, 2272,
		2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358, 5894,
		6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899, 15289,
		16818, 18500, 20350, 22358, 24623, 27086, 29794, 32767};

/* Cabealho adicional para arquivos WAVE que esto em IMA ADPCM */
typedef struct {
	unsigned short int cbSize;
	unsigned short int samplesPerBlock;
	char fact[4];
	unsigned int factSize;
	unsigned int totalSamples;
} AdpcmHeader;

/* Buffer para cada palavra a ser colocada no arquivo de sada */
struct word {
	char buf[WORD_SIZE];
	int cur;
	int low;
	int left;
};

/* Quantizador */
int encodeSample(int sample, int stepSize)
{
	int r = 0;

	if (sample < 0)
	{
		r = r | 1 << 3;
		sample = -sample;
	}

	if (sample >= stepSize)
	{
		r = r | 1 << 2;
		sample = sample - stepSize;
	}

	if (sample >= stepSize/2)
	{
		r = r | 1 << 1;
		sample = sample - stepSize/2;
	}

	if (sample >= stepSize/4)
	{
		r = r | 1;
		sample = sample - stepSize/4;
	}

	return r;
}

/* Funo auxiliar para escrever o cabealho WAVE de arquivos IMA ADPCM */
void outputWaveADPCMHeader(FILE *output, WaveHeader wh, AdpcmHeader ah)
{
	writeBytes(output, wh.riff, 4);
	writeLittleEndianBytes(output, wh.chunkSize, 4);
	writeBytes(output, wh.wave, 4);
	writeBytes(output, wh.fmt, 4);
	writeLittleEndianBytes(output, wh.subchunk1Size, 4);
	writeLittleEndianBytes(output, wh.audioFormat, 2);
	writeLittleEndianBytes(output, wh.numChannels, 2);
	writeLittleEndianBytes(output, wh.sampleRate, 4);
	writeLittleEndianBytes(output, wh.byteRate, 4);
	writeLittleEndianBytes(output, wh.blockAlign, 2);
	writeLittleEndianBytes(output, wh.bitsPerSample, 2);
	writeLittleEndianBytes(output, ah.cbSize, 2);	/* Tamanho extra */
	writeLittleEndianBytes(output, ah.samplesPerBlock, 2);	/* Samples por block */
	writeBytes(output, ah.fact, 4);
	writeLittleEndianBytes(output, ah.factSize, 4);	/* Tamanho do chunk */
	writeLittleEndianBytes(output, ah.totalSamples, 4);	/* Tempo em samples */
	writeBytes(output, wh.data, 4);
	writeLittleEndianBytes(output, wh.subchunk2Size, 4);
}

/* Funo que calcula os valores e escreve o cabealho de arquivos WAVE */
void writeWaveHeader(FILE *output, WaveHeader wh)
{
	AdpcmHeader ah;

	ah.totalSamples = calculateSamplesNumber(wh);

	wh.bitsPerSample = 4;
	wh.subchunk1Size = wh.subchunk1Size + 4;
	wh.audioFormat = 17; /* Cdigo do IMA ADPCM */
	wh.blockAlign = BLOCK_SIZE;
	wh.byteRate = calculateByteRate(wh);

	ah.cbSize = 2;
	strcpy(ah.fact, "fact");
	ah.factSize = 4;
	ah.samplesPerBlock = 1 + ((wh.blockAlign - (4 * wh.numChannels)) * 8) / (wh.bitsPerSample * wh.numChannels);

	wh.subchunk2Size = (4 * wh.numChannels * ah.totalSamples / wh.blockAlign) + (wh.bitsPerSample * ah.totalSamples * wh.numChannels) / 8;
	wh.chunkSize = calculateChunkSize(wh) + 12; /* Tamanho do fact em bytes = 12 */

	outputWaveADPCMHeader(output, wh, ah);
}

/* Escreve um sample no buffer, quando o buffer est cheio, escreve a palavra no arquivo de sada */
void writeSample(FILE *output, struct word *buffer, int quantized)
{
	int i = 0;

	if (buffer->low)
	{
		buffer->buf[buffer->cur] = buffer->buf[buffer->cur] | quantized << 4;
		buffer->low = 0;
		buffer->cur++;
	}
	else
	{
		buffer->buf[buffer->cur] = quantized;
		buffer->low = 1;
	}

	if (buffer->cur >= WORD_SIZE)
	{
		for (i = 0; i < WORD_SIZE; i++)
			fputc(buffer->buf[i], output);
		buffer->left--;
		buffer->cur = 0;
	}
}

/* Fora a escrita do buffer na sada */
void flushBuffer(FILE *output, struct word *buffer)
{
	int i = 0;
	for (i = 0; i < WORD_SIZE; i++)
		fputc(buffer->buf[i], output);
	buffer->cur = 0;
}

/* Funo de desquantizao baseada na funo descrita em:
 * Multimedia Data Standards Update April 15, 1994  Page 35 of 74 */
int predictSample(int sample, int quantized, int step)
{
    int diff = 0;
	if (quantized & 4)
		diff = diff + step;
	if (quantized & 2)
		diff = diff + (step >> 1);
	if (quantized & 1)
		diff = diff + (step >> 2);
	diff = diff + (step >> 3);
	if (quantized & 8)
		diff = -diff;

	int r = sample + diff;
	if (r < -32768) r = -32768;
	else if (r > 32767) r = 32767;

	return r;
}

void encode(FILE *input, FILE *output, WaveHeader wh)
{
	int i, j;
	int numSamples = calculateSamplesNumber(wh);
	int bytesPerSample = wh.bitsPerSample / 8;
	int sample;
	int stepIndex[wh.numChannels];
	int previousSample[wh.numChannels];
	struct word buffer[wh.numChannels];

	for (i = 0; i < wh.numChannels; i++)
	{
		stepIndex[i] = 0;
		buffer[i].cur = 0;
		buffer[i].low = 0;
		buffer[i].left = 0;
	}

	for (i = 0; i < numSamples; i++)
	{
		for (j = 0; j < wh.numChannels; j++)
		{
			/* Lemos a amostra */
			sample = readLittleEndianBytes(input, bytesPerSample);

			/* Se  a primeira do bloco */
			if (buffer[j].left == 0)
			{
				/* Escrevemos o cabealho do bloco */
				writeLittleEndianBytes(output, sample, bytesPerSample);
				previousSample[j] = twoComplementToInt(sample, bytesPerSample);

				fputc(stepIndex[j], output);
				fputc(0, output);

				/* Calculamos quantos valores vo ser escritos ainda */
				buffer[j].left = -1 + BLOCK_SIZE/8;
			}
			else
			{
				/* Convertemos de complemento de 2 para inteiro usual */
				sample = twoComplementToInt(sample, bytesPerSample);

				/* Calculamos a diferena para a amostra anterior */
				sample = sample - previousSample[j];

				/* Quantizamos */
				int quantized = encodeSample(sample, stepSize[stepIndex[j]]);

				/* Desquantizamos o valor da amostra que utilizaremos na iterao seguinte */
				previousSample[j] = predictSample(previousSample[j], quantized, stepSize[stepIndex[j]]);

				/* Ajustamos o passo */
				stepIndex[j] += indexAdjustment[quantized];
				if (stepIndex[j] < 0) stepIndex[j] = 0;
				else if (stepIndex[j] > 88) stepIndex[j] = 88;

				/* Escrevemos a amostra no buffer */
				writeSample(output, &buffer[j], quantized);
			}
		}
	}

	for (i = 0; i < wh.numChannels; i++)
		flushBuffer(output, &buffer[i]);
}

int main(int argc, char *argv[])
{
	FILE *input = NULL, *output = NULL;
	WaveHeader wh;

	wh = openWave(&input, argv[1]);

	/* TODO checar propriedades do wave */

	char temp[100];
	strcpy(temp, argv[1]);
	strcat(temp, ".adpcm.wav");
	output = openFile(temp, "wb");

	writeWaveHeader(output, wh);
	encode(input, output, wh);

	fclose(input);
	fclose(output);

	return 0;
}
