/* MAC0499 - Trabalho de Formatura Supervisionado - Compresso de udio Digital
 * Marcio Masaki Tomiyoshi */

#include "io.h"

/* Abre um arquivo */
FILE* openFile(char *filename, char *mode)
{
	FILE *input = NULL;

	input = fopen(filename, mode);
	if (input == NULL)
		showError("File could not be opened");

	return input;
}

/* Abre um arquivo WAVE e devolve um struct com os valores do cabealho */
WaveHeader openWave(FILE **input, char *filename)
{
	WaveHeader wh;

	*input = openFile(filename, "rb");

	wh = createWaveHeader(*input);
	if (checkWaveHeader(wh) == ERROR)
		showError("Input file format should be WAVE/PCM");

	#if DEBUG != 0
	printWaveHeader(stdout, wh);
	#endif
	checkWaveHeaderConsistency(wh);

	return wh;
}

/* Escreve um inteiro dado com n bits */
void writeInteger(FILE *output, int value, int nBits)
{
	int i = 0;

	while (i < nBits)
		writeBit(output, value & (1 << i++));
}

/* L um inteiro de n bits */
int readInteger(FILE *input, int nBits)
{
	int i = 0, r = 0;

	while (i < nBits)
		r = r | (readBit(input) << i++);
	/*fprintf(stderr,"\n%d\n", r);*/
	return r;
}

/* Escreve n bytes */
void writeBytes(FILE *output, char *str, int nBytes)
{
	fwrite(str, 1, nBytes, output);
}

/* L n bytes */
void readBytes(FILE *input, char *output, int nBytes)
{
	fread(output, 1, nBytes, input);
}

/* L n bytes em formato little endian e transforma adequadamente */
unsigned int readLittleEndianBytes(FILE *input, int nBytes)
{
	unsigned int i, r = 0, shift = 0;
	unsigned char c;

	for (i = 0; i < nBytes; i++)
	{
		c = fgetc(input);
		r = r | ((c & 0x000000FF) << shift);
		shift += 8;
	}

	return r;
}

/* Escreve n bytes no formato little endian. Funo escrita pelo professor
 * Marcelo Queiroz para seu curso de MAC2166 */
void writeLittleEndianBytes(FILE *output, int value, int nBytes) {
	int i;

	for (i = 0; i < nBytes; i++)
	{
		fprintf(output, "%c", value % 256);
		value = value / 256;
	}
}

/* Transforma um valor em complemento de dois para um inteiro usual */
int twoComplementToInt(int value, int nBytes)
{
	const long long int offset = (1 << (nBytes * 8));
	const long long int level = ((1 << ((nBytes * 8) - 1)) - 1);

	if (value > level)
		return value - offset;

	return value;
}

/* Transforma um inteiro usual em notao complemento de dois */
int intToTwoComplement(int value, int nBytes)
{
	const long int offset = (1 << (nBytes * 8));

	if (value < 0)
		return value + offset;

	return value;
}

/* Funo auxiliar para escrita de bits */
int writeBitAux(FILE *output, int bit, int flush)
{
	static int shift = 0;
	static char byte = 0;

	#if DEBUG == 3
	printf("%d", bit && 1);
	#else
	if (bit)
		byte = (byte | (1 << shift));
	shift++;

	if (shift >= 8 || (flush && shift != 1))
	{
		fputc(byte, output);
		shift = 0;
		byte = 0;
	}
	#endif

	return shift;
}

/* Escreve um bit */
int writeBit(FILE *output, int bit)
{
	return writeBitAux(output, bit && 1, 0);
}

/* Fora a escrita do buffer de bits na sada */
int flushBit(FILE *output)
{
	return writeBitAux(output, 0, 1);
}

/* L um bit da entrada */
int readBit(FILE *input)
{
	static int shift = 8;
	static char byte = 0;

	if (shift >= 8)
	{
		byte = fgetc(input);
		shift = 0;
	}

	return (byte & (1 << shift++)) && 1;
}
