/* MAC0499 - Trabalho de Formatura Supervisionado - Compresso de udio Digital
 * Marcio Masaki Tomiyoshi */

#include <stdio.h>
#include <math.h>
#include <string.h>

#include "io.h"

/* u-law */

#define M 255

int encodeSample(double sample)
{
	double y = 0;
	const double c = ((double)127.0 / log(1 + M));

	if (sample >= 0)
		y = (double)255 - c * log(1 + M*(sample));
	else
		y = (double)127 - c * log(1 + M*(-sample));

	if (y - floor(y) > 0.5) /* return int floor(y + 0.5) */
		return (int)ceil(y);

	return (int)y;
}

void writeHeader(FILE *output, WaveHeader wh)
{
	wh.subchunk2Size = calculateSamplesNumber(wh) * wh.numChannels /* * 8/8 */;
	wh.audioFormat = 7;
	wh.bitsPerSample = 8;
	wh.blockAlign = calculateBlockAlign(wh);
	wh.byteRate = calculateByteRate(wh);
	wh.chunkSize = calculateChunkSize(wh);
	outputWaveHeader(output, wh);
}

double normalize(int sample, int maxValue)
{
	return (double)sample/maxValue;
}

void encode(FILE *input, FILE *output, WaveHeader wh)
{
	int i = 0;
	int numSamples = calculateSamplesNumber(wh) * wh.numChannels;
	int bytesPerSample = wh.bitsPerSample / 8;
	int sample;
	int maxValue = 1 << (wh.bitsPerSample - 1);

	for (i = 0; i < numSamples; i++)
	{
		sample = readWaveSample(input, bytesPerSample);

		sample = encodeSample(normalize(sample, maxValue));

		writeLittleEndianBytes(output, sample, 1);
	}
}

int main(int argc, char *argv[])
{
	/*
	 * abreWave
	 * 	se 8 bits/sample fim
	 * leSample
	 * normalizaEntrada
	 * calculaFormula
	 */

	FILE *input = NULL, *output = NULL;
	WaveHeader wh;

	wh = openWave(&input, argv[1]);

	char temp[M];
	strcpy(temp, argv[1]);
	strcat(temp, ".ulaw.wav");
	output = openFile(temp, "wb");

	if (wh.bitsPerSample <= 8)
		showError("Arquivo j possui 8 bits por sample.");

	writeHeader(output, wh);
	encode(input, output, wh);

	fclose(input);
	fclose(output);

	return 0;
}
