/* MAC0499 - Trabalho de Formatura Supervisionado - Compresso de udio Digital
 * Marcio Masaki Tomiyoshi */

#include "wave.h"

/* L a entrada tratando seus bytes como um cabealho WAVE */
WaveHeader createWaveHeader(FILE *input)
{
	WaveHeader wh;

	readBytes(input, wh.riff, 4);
	wh.chunkSize = readLittleEndianBytes(input, 4);
	readBytes(input, wh.wave, 4);
	readBytes(input, wh.fmt, 4);
	wh.subchunk1Size = readLittleEndianBytes(input, 4);
	wh.audioFormat = readLittleEndianBytes(input, 2);
	wh.numChannels = readLittleEndianBytes(input, 2);
	wh. sampleRate = readLittleEndianBytes(input, 4);
	wh.byteRate = readLittleEndianBytes(input, 4);
	wh.blockAlign = readLittleEndianBytes(input, 2);
	wh.bitsPerSample = readLittleEndianBytes(input, 2);
	readBytes(input, wh.data, 4);
	wh.subchunk2Size = readLittleEndianBytes(input, 4);

	return wh;
}

/* Copia 44 bytes (tamanho do cabealho WAVE PCM) da entrada para a sada */
void copyWaveHeader(FILE *input, FILE *output)
{
	char temp[44];
	fpos_t atual;

	fgetpos(input, &atual);
	rewind(input);
	readBytes(input, temp, 44);
	fsetpos(input, &atual);
	writeBytes(output, temp, 44);
}

/* Checa se o cabealho  de um WAVE PCM */
int checkWaveHeader(WaveHeader wh)
{
	if (strncmp((char *)wh.riff, "RIFF", 4) == 0 &&
		strncmp((char *)wh.wave, "WAVE", 4) == 0 &&
		strncmp((char *)wh.fmt, "fmt ", 4) == 0 &&
		strncmp((char *)wh.data, "data", 4) == 0 &&
		wh.audioFormat == 1 &&
		wh.subchunk1Size == 16)
		return OK;
	return ERROR;
}

/* Testa a consistncia de alguns campos WAVE PCM */
int checkWaveHeaderConsistency(WaveHeader wh)
{
	int r = OK;

	if (wh.chunkSize != calculateChunkSize(wh))
	{
		showWarning("Invalid chunk size.");
		r = ERROR;
	}
	if (wh.byteRate != calculateByteRate(wh))
	{
		showWarning("Invalid byte rate.");
		r = ERROR;
	}
	if (wh.blockAlign != calculateBlockAlign(wh))
	{
		showWarning("Invalid block align");
		r = ERROR;
	}

	return r;
}

/* Funes auxiliares para calcular campos do cabealho WAVE */
unsigned int calculateChunkSize(WaveHeader wh)
{
	return 4 + (8 + wh.subchunk1Size) + (8 + wh.subchunk2Size);
}

unsigned int calculateByteRate(WaveHeader wh)
{
	return (wh.sampleRate * wh.numChannels * wh.bitsPerSample / 8);
}

unsigned short int calculateBlockAlign(WaveHeader wh)
{
	return (wh.numChannels * wh.bitsPerSample / 8);
}

unsigned int calculateSamplesNumber(WaveHeader wh)
{
	return 8 * wh.subchunk2Size / (wh.numChannels * wh.bitsPerSample);
}

/* Imprime na sada os dados da struct WaveHeader */
void printWaveHeader(FILE *output, WaveHeader wh)
{
	fprintf(output, "Riff:		%4.4s\n", wh.riff);
	fprintf(output, "Chunk Size:	%u\n", wh.chunkSize);
	fprintf(output, "Wave:		%4.4s\n", wh.wave);
	fprintf(output, "Fmt:		%4.4s\n", wh.fmt);
	fprintf(output, "Subchunk1 Size:	%u\n", wh.subchunk1Size);
	fprintf(output, "Audio Format:	%hu\n", wh.audioFormat);
	fprintf(output, "NumChannels:	%hu\n", wh.numChannels);
	fprintf(output, "Sample Rate:	%u\n", wh.sampleRate);
	fprintf(output, "Byte Rate:	%u\n", wh.byteRate);
	fprintf(output, "Block Align:	%hu\n", wh.blockAlign);
	fprintf(output, "Bits/Sample:	%hu\n", wh.bitsPerSample);
	fprintf(output, "Data:		%4.4s\n", wh.data);
	fprintf(output, "Subchunk2 Size:	%u\n", wh.subchunk2Size);
	fprintf(output, "NumSamples:	%u\n", calculateSamplesNumber(wh));
}

/* Escreve na sada o cabealho Wave */
void outputWaveHeader(FILE *output, WaveHeader wh)
{
	writeBytes(output, wh.riff, 4);
	writeLittleEndianBytes(output, wh.chunkSize, 4);
	writeBytes(output, wh.wave, 4);
	writeBytes(output, wh.fmt, 4);
	writeLittleEndianBytes(output, wh.subchunk1Size, 4);
	writeLittleEndianBytes(output, wh.audioFormat, 2);
	writeLittleEndianBytes(output, wh.numChannels, 2);
	writeLittleEndianBytes(output, wh.sampleRate, 4);
	writeLittleEndianBytes(output, wh.byteRate, 4);
	writeLittleEndianBytes(output, wh.blockAlign, 2);
	writeLittleEndianBytes(output, wh.bitsPerSample, 2);
	writeBytes(output, wh.data, 4);
	writeLittleEndianBytes(output, wh.subchunk2Size, 4);
}

/* L uma amostra de n bytes de um arquivo WAVE PCM */
int readWaveSample(FILE *input, int nBytes)
{
	int r = readLittleEndianBytes(input, nBytes);

	if (nBytes == 1)
		return r - 128;

	r = twoComplementToInt(r, nBytes);

	return r;
}

/* Escreve uma amostra na sada com n bytes */
void writeWaveSample(FILE *output, int value, int nBytes)
{
	int r;

	if (nBytes == 1)
		writeLittleEndianBytes(output, value + 128, nBytes);
	else
	{
		r = intToTwoComplement(value, nBytes);
		writeLittleEndianBytes(output, r, nBytes);
	}
}
